<?php
/**
 * Feed retrieval, parsing, and manipulation
 */
class Feed {

    /** @var string The XML namespace for Atom feeds */
    public const string ATOM_NS = 'http://www.w3.org/2005/Atom';

    /** @var string The XML namespace for the `<content:encoded>` tag that allows HTML content in a feed */
    public const string CONTENT_NS = 'http://purl.org/rss/1.0/modules/content/';

    /**
     * When parsing XML into a DOMDocument, errors are presented as warnings; this creates an exception for them
     *
     * @param int $errno The error level encountered
     * @param string $errstr The text of the error encountered
     * @return bool False, to delegate to the next error handler in the chain
     * @throws DOMException If the error is a warning
     */
    private static function xmlParseError(int $errno, string $errstr): bool {
        if ($errno == E_WARNING && substr_count($errstr, 'DOMDocument::loadXml()') > 0) {
            throw new DOMException($errstr, $errno);
        }
        return false;
    }

    /**
     * Parse a feed into an XML tree
     *
     * @param string $content The feed's RSS content
     * @return array|DOMDocument[]|string[] ['ok' => feed] if successful, ['error' => message] if not
     */
    public static function parseFeed(string $content): array {
        set_error_handler(self::xmlParseError(...));
        try {
            $feed = new DOMDocument();
            $feed->loadXML($content);
            return ['ok' => $feed];
        } catch (DOMException $ex) {
            return ['error' => $ex->getMessage()];
        } finally {
            restore_error_handler();
        }
    }

    /**
     * Get the value of a child element by its tag name
     *
     * @param DOMElement $element The parent element
     * @param string $tagName The name of the tag whose value should be obtained
     * @return string The value of the element (or "[element] not found" if that element does not exist)
     */
    private static function eltValue(DOMElement $element, string $tagName): string {
        $tags = $element->getElementsByTagName($tagName);
        return $tags->length == 0 ? "$tagName not found" : $tags->item(0)->textContent;
    }

    /**
     * Retrieve the feed
     *
     * @param string $url
     * @return array|DOMDocument[]|string[]|DOMElement[]
     *      ['ok' => feedXml, 'url' => actualUrl, 'channel' => channel, 'updated' => updatedDate] if successful,
     *      ['error' => message] if not
     */
    public static function retrieveFeed(string $url): array {
        $feedReq = curl_init($url);
        curl_setopt($feedReq, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($feedReq, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($feedReq, CURLOPT_CONNECTTIMEOUT, 5);
        curl_setopt($feedReq, CURLOPT_TIMEOUT,        15);

        $feedContent = curl_exec($feedReq);

        $result = array();
        $error  = curl_error($feedReq);
        $code   = curl_getinfo($feedReq, CURLINFO_RESPONSE_CODE);
        if ($error) {
            $result['error'] = $error;
        } else if ($code == 200) {
            $parsed = self::parseFeed($feedContent);
            if (array_key_exists('error', $parsed)) {
                $result['error'] = $parsed['error'];
            } else {
                $result['ok']  = $parsed['ok'];
                $result['url'] = curl_getinfo($feedReq, CURLINFO_EFFECTIVE_URL);

                $channel = $result['ok']->getElementsByTagName('channel')->item(0);
                if ($channel instanceof DOMElement) {
                    $result['channel'] = $channel;
                } else {
                    return ['error' => "Channel element not found ($channel->nodeType)"];
                }

                // In Atom feeds, lastBuildDate contains the last time an item in the feed was updated; if that is not
                // present, use the pubDate element instead
                $updated = self::eltValue($channel, 'lastBuildDate');
                if ($updated == 'lastBuildDate not found') {
                    $updated = self::eltValue($channel, 'pubDate');
                    if ($updated == 'pubDate not found') $updated = null;
                }
                $result['updated'] = Data::formatDate($updated);
                return $result;

            }
        } else {
            $result['error'] = "Prospective feed URL $url returned HTTP Code $code: $feedContent";
        }

        curl_close($feedReq);
        return $result;
    }

    /**
     * Extract the fields we need to keep from the feed
     *
     * @param DOMElement $item The item from the feed
     * @return array The fields for the item as an associative array
     */
    private static function itemFields(DOMElement $item): array {
        $itemGuid = self::eltValue($item, 'guid');
        $updNodes = $item->getElementsByTagNameNS(self::ATOM_NS,    'updated');
        $encNodes = $item->getElementsByTagNameNS(self::CONTENT_NS, 'encoded');
        return [
            'guid'      => $itemGuid == 'guid not found' ? self::eltValue($item, 'link') : $itemGuid,
            'title'     => self::eltValue($item, 'title'),
            'link'      => self::eltValue($item, 'link'),
            'published' => Data::formatDate(self::eltValue($item, 'pubDate')),
            'updated'   => Data::formatDate($updNodes->length > 0 ? $updNodes->item(0)->textContent : null),
            'content'   => $encNodes->length > 0 ? $encNodes->item(0)->textContent
                                                 : self::eltValue($item, 'description'),
            'isEncoded' => $encNodes->length > 0
        ];
    }

    /**
     * Update a feed item
     *
     * @param int $itemId The ID of the item to be updated
     * @param array $item The fields from the updated item
     * @param SQLite3 $db A database connection to use for the update
     */
    private static function updateItem(int $itemId, array $item, SQLite3 $db): void {
        $query = $db->prepare(<<<'SQL'
            UPDATE item
               SET title        = :title,
                   published_on = :published,
                   updated_on   = :updated,
                   content      = :content,
                   is_encoded   = :encoded,
                   is_read      = 0
             WHERE id = :id
            SQL);
        $query->bindValue(':title',     $item['title']);
        $query->bindValue(':published', $item['published']);
        $query->bindValue(':updated',   $item['updated']);
        $query->bindValue(':content',   $item['content']);
        $query->bindValue(':encoded',   $item['isEncoded']);
        $query->bindValue(':id',        $itemId);
        $query->execute();
    }

    /**
     * Add a feed item
     *
     * @param int $feedId The ID of the feed to which the item should be added
     * @param array $item The fields for the item
     * @param SQLite3 $db A database connection to use for the addition
     */
    private static function addItem(int $feedId, array $item, SQLite3 $db): void {
        $query = $db->prepare(<<<'SQL'
            INSERT INTO item (
                feed_id, item_guid, item_link, title, published_on, updated_on, content, is_encoded
            ) VALUES (
                :feed, :guid, :link, :title, :published, :updated, :content, :encoded
            )
            SQL);
        $query->bindValue(':feed',      $feedId);
        $query->bindValue(':guid',      $item['guid']);
        $query->bindValue(':link',      $item['link']);
        $query->bindValue(':title',     $item['title']);
        $query->bindValue(':published', $item['published']);
        $query->bindValue(':updated',   $item['updated']);
        $query->bindValue(':content',   $item['content']);
        $query->bindValue(':encoded',   $item['isEncoded']);
        $query->execute();
    }

    /**
     * Update a feed's items
     *
     * @param int $feedId The ID of the feed to which these items belong
     * @param DOMElement $channel The RSS feed items
     * @return array ['ok' => true] if successful, ['error' => message] if not
     */
    public static function updateItems(int $feedId, DOMElement $channel, SQLite3 $db): array {
        try {
            foreach ($channel->getElementsByTagName('item') as $rawItem) {
                $item = self::itemFields($rawItem);
                $existsQuery = $db->prepare(
                    'SELECT id, published_on, updated_on FROM item WHERE feed_id = :feed AND item_guid = :guid');
                $existsQuery->bindValue(':feed', $feedId);
                $existsQuery->bindValue(':guid', $item['guid']);
                $exists = $existsQuery->execute();
                if ($exists) {
                    $existing = $exists->fetchArray(SQLITE3_ASSOC);
                    if ($existing) {
                        if (   $existing['published_on']     != $item['published']
                            || $existing['updated_on'] ?? '' != $item['updated'] ?? '') {
                            self::updateItem($existing['id'], $item, $db);
                        }
                    } else {
                        self::addItem($feedId, $item, $db);
                    }
                } else {
                    throw new Exception($db->lastErrorMsg());
                }
            }
        } catch (Exception $ex) {
            return ['error' => $ex->getMessage()];
        }
        return ['ok', true];
    }

    /**
     * Refresh a feed
     *
     * @param string $url The URL of the feed to be refreshed
     * @param SQLite3 $db A database connection to use to refresh the feed
     * @return array|string[]|true[] ['ok' => true] if successful, ['error' => message] if not
     */
    private static function refreshFeed(string $url, SQLite3 $db): array {
        $feedQuery = $db->prepare('SELECT id FROM feed WHERE url = :url AND user_id = :user');
        $feedQuery->bindValue(':url',  $url);
        $feedQuery->bindValue(':user', $_REQUEST[Key::USER_ID]);
        $feedResult = $feedQuery->execute();
        $feedId = $feedResult ? $feedResult->fetchArray(SQLITE3_NUM)[0] : -1;
        if ($feedId < 0) return ['error' => "No feed for URL $url found"];

        $feed = self::retrieveFeed($url);
        if (array_key_exists('error', $feed)) return $feed;

        $itemUpdate = self::updateItems($feedId, $feed['channel'], $db);
        if (array_key_exists('error', $itemUpdate)) return $itemUpdate;

        $urlUpdate = $url == $feed['url'] ? '' : ', url = :url';
        $feedUpdate = $db->prepare(<<<SQL
            UPDATE feed
               SET title      = :title,
                   updated_on = :updated,
                   checked_on = :checked
                   $urlUpdate
             WHERE id = :id
            SQL);
        $feedUpdate->bindValue(':title',   self::eltValue($feed['channel'], 'title'));
        $feedUpdate->bindValue(':updated', $feed['updated']);
        $feedUpdate->bindValue(':checked', Data::formatDate('now'));
        $feedUpdate->bindValue(':id',      $feedId);
        if ($urlUpdate != '') $feedUpdate->bindValue(':url', $feed['url']);
        $feedUpdate->execute();

        return ['ok' => true];
    }

    /**
     * Add an RSS feed
     *
     * @param string $url The URL of the RSS feed to add
     * @return array ['ok' => feedId] if successful, ['error' => message] if not
     */
    public static function add(string $url, SQLite3 $db): array {
        $feed = self::retrieveFeed($url);
        if (array_key_exists('error', $feed)) return $feed;

        $query = $db->prepare(<<<'SQL'
            INSERT INTO feed (user_id, url, title, updated_on, checked_on)
            VALUES (:user, :url, :title, :updated, :checked)
            SQL);
        $query->bindValue(':user',    $_REQUEST[Key::USER_ID]);
        $query->bindValue(':url',     $feed['url']);
        $query->bindValue(':title',   self::eltValue($feed['channel'], 'title'));
        $query->bindValue(':updated', $feed['updated']);
        $query->bindValue(':checked', Data::formatDate('now'));
        $result = $query->execute();

        $feedId = $result ? $db->lastInsertRowID() : -1;
        if ($feedId < 0) return ['error' => $db->lastErrorMsg()];

        $result = self::updateItems($feedId, $feed['channel'], $db);
        if (array_key_exists('error', $result)) return $result;

        return ['ok' => $feedId];
    }

    /**
     * Update an RSS feed
     *
     * @param array $existing The existing RSS feed
     * @param string $url The URL with which the existing feed should be modified
     * @return bool[]|string[] [ 'ok' => true ] if successful, [ 'error' => message ] if not
     */
    public static function update(array $existing, string $url, SQLite3 $db): array {
        $query = $db->prepare('UPDATE feed SET url = :url WHERE id = :id AND user_id = :user');
        $query->bindValue(':url',  $url);
        $query->bindValue(':id',   $existing['id']);
        $query->bindValue(':user', $_REQUEST[Key::USER_ID]);
        $query->execute();

        return self::refreshFeed($url, $db);
    }

    /**
     * @param SQLite3 $db
     * @return array|true[] ['ok => true] if successful, ['error' => message] if not (may have multiple error lines)
     */
    public static function refreshAll(SQLite3 $db): array {
        $query = $db->prepare('SELECT url FROM feed WHERE user_id = :user');
        $query->bindValue(':user', $_REQUEST[Key::USER_ID]);
        $result = $query->execute();
        $url    = $result ? $result->fetchArray(SQLITE3_NUM) : false;
        if ($url) {
            $errors = array();
            while ($url) {
                $updateResult = self::refreshFeed($url[0], $db);
                if (array_key_exists('error', $updateResult)) $errors[] = $updateResult['error'];
                $url = $result->fetchArray(SQLITE3_NUM);
            }
            return sizeof($errors) == 0 ? ['ok' => true] : ['error' => implode("\n", $errors)];
        }
        return ['error' => $db->lastErrorMsg()];
    }
}
